require 'rake'

# @author Alex Bezobchuk
module ThinIce
  extend Rake::DSL

  # A custom Exception to be thrown when an undesired exit status
  # is returned when spawning a process.
  class ThinCommandFailure < StandardError; end

  # Forks a process given a list of arguments and options
  # @param args [String] list of command arguments and options
  # @return [Integer] the resulting exit status
  # @raise [ThinCommandFailure] if the exit status of the process
  #   is non-zero
  def self.invoke(args)
    pid = Process.spawn(*args)
    _, result = Process.wait2(pid)

    s = result.exitstatus
    if s != 0
      raise ThinCommandFailure, "thin command exited with status #{s}"
    end
  end

  # ThinManager is designed to help construct the exact command to be excuted
  # by the Rake task.
  class ThinManager

    # @!attribute env [String] Environment
    # @!attribute config_path [String] Path to YML configuration file
    # @!attribute thin_path [String] Path to Thin executable
    # @!attribute rackup_path [String] Path to Rackup configuration
    attr_reader :env, :config_path, :thin_path, :rackup_path

    # Creates a [ThinManager] object that is used to to aid rake task
    #   generation
    # @param options [Hash] Hash containing various configuration
    #   options
    # @return [ThinManager] ThinManager object
    def initialize(options = {})
      @env             = options[:env]
      @thin_path       = options[:thin_path]
      @config_path     = options[:config_path]
      @rackup_path     = options[:rackup_path]
    end

    # @return [Array] An Array cotaining the environment string. May be
    #   empty
    def env_param
      (env) ? [env] : []
    end

    # @return [Array] An Array cotaining the Thin executable location. Uses
    #   local Thin gem by default
    def thin_exec_param
      [thin_path || 'thin']
    end

    # @return [Array] An Array cotaining the path to a YML configuration file.
    #   Empty by default
    def config_path_param
      (config_path) ? ['-C', config_path] : []
    end

    # @return [Array] An Array cotaining the path to a Rackup configuration file.
    #   Empty by default
    def rackup_path_param
      (rackup_path) ? ['-R', rackup_path] : []
    end

    # @return [String] A concatination of strings resulting in the Thin
    #   start command to be invoked by Rake
    def start_cmd
      env_param + thin_exec_param + rackup_path_param + config_path_param + ['start']
    end

    # @return [String] A concatination of strings resulting in the Thin
    #   stop command to be invoked by Rake
    def stop_cmd
      env_param + thin_exec_param + rackup_path_param + config_path_param + ['stop']
    end

  end

  # Generates a set of Rake tasks to be used within a specific namespace
  #   that handle starting and stoping a Thin server instance
  # @param options [Hash] Hash containing various configuration
  #   options
  def self.create_tasks(options = {})

    thin_manager = ThinManager.new(options)

    desc 'Start the Thin server'
    task :start do
      invoke thin_manager.start_cmd
    end

    desc 'Stop the Thin server'
    task :stop do
      invoke thin_manager.stop_cmd
    end

    desc 'Restart the Thin server'
    task :restart do
      invoke thin_manager.stop_cmd
      invoke thin_manager.start_cmd
    end
  end
end
